
/*
	Module: Reading Values from Serial Analog/Digital Converter 
	Copyright (C) 1999/2000  A. Schiffler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	To contact the author try:
	Andreas Schiffler                    aschiffler@home.com
	Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo
	4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada
	+1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)

*/

#include "adcread.h"

/* ------- Actual serial command/result exchange  */

float readVoltage(struct adcreader *adc, int channel) 
{
 float value;

 value=0.0;

 return(value);
}

/* ------- Serial A/D-converter reader thread */

int ADCReaderThread(struct adcreader *adc)
{
 int inloop;
 int channel;
 float value;
 int localfd;
 int bytesread;
 int result;
 int repeatcount;
 unsigned char buf[2];

 /* Get local fd handle */
 adcreader_lock(adc);
 localfd=adc->fd;
 adcreader_unlock(adc);
 /* Go into processing loop */
 inloop=1;
 while (inloop) {
  for (channel=0; channel<NUM_CHANNELS; channel++) {
   /* Write control byte */
   write(localfd,&controlbytes[channel],1);
   /* Read 2 results bytes */
   fcntl(localfd, F_SETFL, FNDELAY);
   bytesread=0;
   repeatcount=0;
   while ((bytesread<2) && (repeatcount<1000000)) {
    result = read(localfd,&buf[bytesread],1);
    if (result>0) {
     bytesread += result;
    } else {
     repeatcount++;
    }  
   }  
   value=(float)(buf[0]+buf[1]*256)*(1.0/4096.0);
   /* Store value */
   adcreader_lock(adc);
   adc->values[channel]=value;
   adcreader_unlock(adc);
   /* Pause */
   usleep(5000);
  }
 }
 
 return(0);
}

/* -------- Thread locking functions */

int adcreader_lock(struct adcreader *adc)
{
 /* Lock shared variables */ 
 pthread_mutex_lock(&adc->adcread_mutex);
 
 return(0);
}

int adcreader_unlock(struct adcreader *adc)
{
 /* Unlock shared variables */ 
 pthread_mutex_unlock(&adc->adcread_mutex);
 
 return(0);
}

/* ------- Start the reader thread */

int start_adcreader (struct adcreader *adc)
{
 pthread_t reader_thread;

 if (adc->enabled) {
  /* Start thread */
  if (pthread_create(&reader_thread, NULL,(void *)ADCReaderThread,(void *)adc)==-1) {
   perror ("pthread_create()");
   return(-1);
  } else {
   adc->thread=reader_thread;
   return(0);
  }
 } else {
  fprintf (stderr,"start_adcreader: cannot start thread - device not enabled.\n");
  return(-1);
 } 
}

/* ----- Initialize the device and prepare thread */

int init_adcreader(struct adcreader *adc, char* device, int type)
{
 int channel;
 struct termios newtio; 

 /* Turn off */
 adc->enabled=0;

 /* Initialize variables */
 adc->thread=-1;
 for (channel=0; channel<NUM_CHANNELS; channel++) {
  adc->values[channel]=0.0;
 }

 /* Initialize sycro mutex */
 if (pthread_mutex_init(&adc->adcread_mutex, NULL)==-1) {
  perror("pthread_mutex_init()");
  return(-1);
 }

 /* Check type */
 if (type==ADC_NONE) {
  return(0);
 }
 
 /* Open serial device */
 if ((adc->fd = open(device, O_RDWR | O_NOCTTY | O_NDELAY ))==-1) {
  perror("open()");
  return(-1);
 }

 tcgetattr(adc->fd,&adc->oldtio); 
 bzero(&newtio, sizeof(newtio)); 

 newtio.c_cflag = B115200 | CRTSCTS | CS8 | CLOCAL | CREAD;
 newtio.c_iflag = IGNPAR;
 newtio.c_oflag = 0;
 newtio.c_lflag = 0;
 newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
 newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */

 tcflush(adc->fd, TCIFLUSH);
 tcsetattr(adc->fd,TCSANOW,&newtio);

 /* Enable and return */ 
 adc->enabled=1;
 return(0);
}

/* ------- Stop reader thread */

int stop_adcreader (struct adcreader *adc)
{
 /* restore the old port settings */
 tcsetattr(adc->fd,TCSANOW,&adc->oldtio); 

 /* Shutdown thread */
 if (adc->thread!=-1) {
  pthread_cancel(adc->thread);
 }

 return(0);
}
