/*

	Program: Beyond Hierarchy - Control Program 
	Copyright (C) 1999/2000  A. Schiffler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	To contact the author try:
	Andreas Schiffler                    aschiffler@home.com
	Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo
	4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada
	+1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)

*/


/* Beyond Hierarchy */

/*
 * NOTE: Requires XWindows running at 16bit depth 565 weight.
 */

#include <stdio.h>
#include <stdlib.h>

#include "xutil.h"
#include "joyread.h"
#include "dvdcontrol.h"
#include "touchscan.h"
#include "lights.h"

static char copyright[]   = "Andreas Schiffler";
static char dateversion[] = "02-03/00, Ver 1.3";

/* Colors and screen definitions */

#define WHITE	0xffff

#define WIDTH	640
#define HEIGHT	480
#define DEPTH 	2
#define SIZE	(HEIGHT*WIDTH*DEPTH)

#define TARGETFPS	18.0

/* ----------- Number of devices - device mapping */

#define NUMADC		2
#define NUMJOY		6
#define NUMDVD		7
#define NUMTOUCH	NUM_SENSORS

/* -------------- Analog/Digital converters for joysticks */

static char *adcdevice[NUMADC] = {	"/dev/ttyC8",
					"/dev/ttyC9"
				};

static int adctype[NUMADC] = {	ADC_ADIO,
				ADC_ADIO
                             };


/* -------------- DVD players */

static char *dvddevice[NUMDVD] = {	"/dev/ttyC0",
					"/dev/ttyC1",
					"/dev/ttyC2",
					"/dev/ttyC3",
					"/dev/ttyC4",
					"/dev/ttyC5",
					"/dev/ttyC6"
				   };

static int dvdtype[NUMDVD] = {	DVD_PIONEER,
				DVD_PIONEER,
				DVD_PIONEER,
				DVD_PIONEER,
				DVD_PIONEER,
				DVD_PIONEER,
				DVD_PIONEER
                              };

/* ----------- Touchbox */

static int touchtype = TOUCH_HAPP;

static char touchdevice[] = "/dev/ttyC10";

/* ----------- Light controller */

static int lighttype = LIGHT_CONRAD;

static char lightdevice[] = "/dev/ttyC11";

static unsigned char touchlight[6] = { 2,4,8,16,32,64 };

static int lightstate = 0;

/* ----------- Filenames */

#define calibration_filename	"joystick_calibrations.txt"
#define matrix_filename		"chapter_mapping.txt"
#define log_filename		"chapter_mapping.log"
#define conversation_filename	"conversation_mapping.txt"
#define log2_filename		"conversation_mapping.log"

/* ----------- Screen layout */

#define XSEGMENTS	4
#define YSEGMENTS	2

#define JOYBORDER	8
#define JOYSEGMENTS	7
#define JOYTEXTLINES	6
#define JOYSEGMENTX	16
#define JOYSEGMENTY	16
#define JOYCENTERX	(((WIDTH/XSEGMENTS)-(JOYSEGMENTS*JOYSEGMENTX)-2*JOYBORDER)/2)
#define JOYMODECOL	7
#define JOYMODEROW	3
#define JOYSELCOL	7
#define JOYSELROW	4
#define JOYLOCKCOL	7
#define JOYLOCKROW	5

/* Joystick-to-Segment mapping */

static int joytarget[JOYSEGMENTS][JOYSEGMENTS] = {
			{  1,  1, -1, -1, -1, 13, 13 },
			{  2,  2, -1, -1, -1, 14, 14 },
			{  3,  8,  9, 10, 11, 12, 15 },
			{  4,  8,  9, 10, 11, 12, 16 },
			{  5,  8,  9, 10, 11, 12, 17 },
			{  6, -1, -1, -1, -1, -1, 18 },
			{  7,  7, -1, -1, -1, 19, 19 }
			};

static int joytargetx[19] = {0,0,0,0,0,0,0, 1,2,3,4,5, 6,6,6,6,6,6,6};
static int joytargety[19] = {0,1,2,3,4,5,6, 3,3,3,3,3, 0,1,2,3,4,5,6};

static char joyname[NUMJOY][3] = { "A1","A2","A3","B1","B2","B3"};
		
/* Screen helper variables */

static int joyoriginx[NUMJOY];
static int joyoriginy[NUMJOY];
static int joycenterx[NUMJOY];
static int joycentery[NUMJOY];
static int joysizex[NUMJOY];
static int joysizey[NUMJOY];

/* Joystick subdivisions */

#define SUBDIVX		4
#define SUBDIVY		5

/* Chapter-Mapping Matrix for Chairs */

#define MATRIXINDEXA	(NUMJOY)
#define MATRIXINDEXB	(3*JOYSEGMENTS-2)
#define MATRIXINDEXC	(SUBDIVY)
#define MATRIXINDEXD	(2)

static int mapping_matrix[MATRIXINDEXA][MATRIXINDEXB][MATRIXINDEXC][MATRIXINDEXD];

/* Conversation-Mapping Matrix for Chairs */

#define MATRIXINDEXE	(6)
#define MATRIXINDEXF	(6)

static int conversation_matrix[MATRIXINDEXE][MATRIXINDEXF];
static double conversation_time[MATRIXINDEXE][MATRIXINDEXF];

/* DVD helper variables */

static int dvdchapter[NUMDVD] = { -1, -1, -1, -1, -1, -1, -1 };
static double dvdcommandtime[NUMDVD] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };

/* Touchbox screen layout */

#define TOUCHBORDER	8
#define TOUCHTEXTLINES	7
#define TOUCHSEGMENTX	32
#define TOUCHSEGMENTY	56
#define TOUCHCENTERX	(((WIDTH/XSEGMENTS)-(3*TOUCHSEGMENTX)-2*TOUCHBORDER)/2)	
#define TOUCHMODECOL	7
#define TOUCHMODEROW	3
#define TOUCHSELCOL	7
#define TOUCHSELROW	4

/* Touchbox helper variables */

static int touchoriginx;
static int touchoriginy;

static int manualtouchstate[NUMTOUCH];
static int touchcenterx[NUMTOUCH];
static int touchcentery[NUMTOUCH];

static int lockjoystick[NUMTOUCH];

/* Title screen layout */

#define TITLEBORDER	8
#define FPSROW		5
#define FPSCOL		8
#define LIGHTROW	8
#define LIGHTCOLUMN	8
#define CHAPTERROW	10
#define CHAPTERCOLUMN	15

/* ================ Subroutines ======= */

/* Load chapter mapping matrix */

void LoadMappingMatrix()
{
 int i,j,k,l;
 FILE *iofile;
 FILE *logfile;
 char aline[256];
 char character[256];
 char dummy[256];
 int indexa;  /* Character */
 int indexb;  /* Segment */
 int indexc;  /* Subsegment */
 int chapter;
 int chapteralternate;
 int dologging;
    
 /* Say something */
 fprintf (stdout,"Loading mapping matrix from '%s'.\n",matrix_filename);
 fprintf (stdout,"Logging process to '%s'.\n",log_filename);
 fflush (stdout);
 
 /* Clear matrix */
 for (i=0; i<MATRIXINDEXA; i++) {
  for (j=0; j<MATRIXINDEXB; j++) {
   for (k=0; k<MATRIXINDEXC; k++) {
    for (l=0; l<MATRIXINDEXD; l++) {
     mapping_matrix[i][j][k][l]=-1;
    }
   }
  }
 }

 /* Open logfile */
 dologging=0;
 if ((logfile=fopen(log_filename,"w"))!=NULL) {
  dologging=1;
 }
  
 /* Load configuration */
 if ((iofile=fopen(matrix_filename,"r"))!=NULL) {
  fgets(aline,256,iofile);
  while (!feof(iofile)) {
   /* Process line */
   if ((aline[0]!='#') || (strlen(aline)>0)) {
    /* Init variables */
    indexa=-1;
    indexb=0;
    indexc=0;
    chapter=1;
    chapteralternate=1;   
    /* Read the line */
    sscanf (aline,"%s %i %i %s %i %i\n",character,&indexb,&indexc,dummy,&chapter,&chapteralternate);
    /* Convert character to index */
    if (strcasecmp(character,"A1")==0) indexa=0;
    if (strcasecmp(character,"A2")==0) indexa=1;
    if (strcasecmp(character,"A3")==0) indexa=2;
    if (strcasecmp(character,"B1")==0) indexa=3;
    if (strcasecmp(character,"B2")==0) indexa=4;
    if (strcasecmp(character,"B3")==0) indexa=5;
    indexb--;
    indexc--;
    /* Sanity check on variables */
    if ((indexa>-1) && (indexa<MATRIXINDEXA) &&
        (indexb>-1) && (indexb<MATRIXINDEXB) &&
        (indexc>-1) && (indexc<MATRIXINDEXC) &&
        (chapter>0) && (chapter<100) &&
        (chapteralternate>0) && (chapteralternate<100)
      ) {
      /* Store chapters */
      mapping_matrix[indexa][indexb][indexc][0]=chapter;
      mapping_matrix[indexa][indexb][indexc][1]=chapteralternate;
      /* Log what we are doing */
      fprintf(logfile,"Character: %2s  Segment: %2i %2i --- Chapters: %2i %2i\n",character,indexb,indexc,chapter,chapteralternate);
    }
   }
   fgets(aline,256,iofile);
  }
  fclose(iofile);
 }
 
 /* Close logfile */
 if (dologging) {
  fclose(logfile);
 }
}

/* Load conversation mapping matrix */

void LoadConversationMatrix()
{
 int i,j;
 FILE *iofile;
 FILE *logfile;
 char aline[256];
 char characterA[256];
 char characterB[256];
 char dummy[256];
 int indexa, indexb;
 int chapter;  /* Character */
 float runningtime;  /* RunningTime */
 int dologging;
     
 /* Say something */
 fprintf (stdout,"Loading conversation matrix from '%s'.\n",conversation_filename);
 fflush (stdout);
 
 /* Clear matrix */
 for (i=0; i<MATRIXINDEXE; i++) {
  for (j=0; j<MATRIXINDEXF; j++) {
   conversation_matrix[i][j]=-1;
  }
 }

 /* Open logfile */
 dologging=0;
 if ((logfile=fopen(log2_filename,"w"))!=NULL) {
  dologging=1;
 }

 /* Load configuration */
 if ((iofile=fopen(conversation_filename,"r"))!=NULL) {
  fgets(aline,256,iofile);
  while (!feof(iofile)) {
   /* Process line */
   if ((aline[0]!='#') || (strlen(aline)>0)) {
    /* Init variables */
    indexa=-1;
    indexb=-1;
    chapter=1;
    runningtime=0.0;
    /* Read the line */
    sscanf (aline,"%s %s %s %i %f\n",characterA,characterB,dummy,&chapter,&runningtime);
    /* Convert character to index */
    if (strcasecmp(characterA,"A1")==0) indexa=0;
    if (strcasecmp(characterA,"A2")==0) indexa=1;
    if (strcasecmp(characterA,"A3")==0) indexa=2;
    if (strcasecmp(characterA,"B1")==0) indexa=3;
    if (strcasecmp(characterA,"B2")==0) indexa=4;
    if (strcasecmp(characterA,"B3")==0) indexa=5;
    if (strcasecmp(characterB,"A1")==0) indexb=0;
    if (strcasecmp(characterB,"A2")==0) indexb=1;
    if (strcasecmp(characterB,"A3")==0) indexb=2;
    if (strcasecmp(characterB,"B1")==0) indexb=3;
    if (strcasecmp(characterB,"B2")==0) indexb=4;
    if (strcasecmp(characterB,"B3")==0) indexb=5;
    /* Sanity check on variables */
    if ((indexa>-1) && (indexa<MATRIXINDEXE) &&
        (indexb>-1) && (indexb<MATRIXINDEXF) &&
        (chapter>0) && (chapter<100) &&
        (runningtime>0.0) && (runningtime<1000.0)
      ) {
      /* Store chapters */
      conversation_matrix[indexa][indexb]=chapter;
      conversation_time[indexa][indexb]=runningtime;
      conversation_matrix[indexb][indexa]=chapter;
      conversation_time[indexb][indexa]=runningtime;
      /* Log what we are doing */
      fprintf(logfile,"Characters: %2s + %2s  --- Chapter: %2i Time: %7.2f\n",characterA,characterB,chapter,runningtime);
    }
   }
   fgets(aline,256,iofile);
  }
  fclose(iofile);
 }
 
 /* Close logfile */
 if (dologging) {
  fclose(logfile);
 }
}

/* Draw joystick background */

void drawJoystickH(struct xwinbuffer *xwin, char *label, int  originx, int originy)
{
 int i,y1,y2,x1,x2;
 
 drawbox (xwin, originx+JOYBORDER, originy+JOYBORDER, originx+WIDTH/XSEGMENTS-2*JOYBORDER, originy+JOYBORDER+8, RGB24RGB16(128,64,64));
 drawstring (xwin, originx+JOYBORDER, originy+JOYBORDER, label, WHITE);
 drawstring (xwin, originx+JOYBORDER, originy+JOYBORDER+JOYMODEROW*8, "Mode:", WHITE);
 drawstring (xwin, originx+JOYBORDER, originy+JOYBORDER+JOYSELROW*8, "Sel.:", WHITE);
 drawstring (xwin, originx+JOYBORDER, originy+JOYBORDER+JOYLOCKROW*8, "Lock:", WHITE);
 for (i=0; i<=JOYSEGMENTS; i++) {
  y1=originy+2*JOYBORDER+JOYTEXTLINES*8+JOYSEGMENTY*i;
  x1=originx+JOYBORDER+JOYCENTERX;
  x2=x1+JOYSEGMENTX;
  hline (xwin,x1,x2,y1,WHITE);
  x1=originx+JOYBORDER+(JOYSEGMENTS-1)*JOYSEGMENTX+JOYCENTERX;
  if ( (i==(JOYSEGMENTS/2)) || (i==(JOYSEGMENTS/2+1)) ) {
   hline (xwin,x2,x1,y1,WHITE);
  } 
  x2=x1+JOYSEGMENTX;
  hline (xwin,x1,x2,y1,WHITE);
 }  
 for (i=0; i<=JOYSEGMENTS; i++) {
  x1=originx+JOYBORDER+JOYSEGMENTX*i+JOYCENTERX;
  if ((i<=1) || (i>=(JOYSEGMENTS-1)) ) {
   y1=originy+2*JOYBORDER+JOYTEXTLINES*8;
   y2=originy+2*JOYBORDER+JOYTEXTLINES*8+JOYSEGMENTY*JOYSEGMENTS;
  } else {  
   y1=originy+2*JOYBORDER+JOYTEXTLINES*8+JOYSEGMENTY*(JOYSEGMENTS/2);
   y2=y1+JOYSEGMENTY;
  } 
  vline (xwin,x1,y1,y2,WHITE);
 }
}

/* Draw touchbox background */

void drawTouchbox(struct xwinbuffer *xwin, char *label, int  originx, int originy)
{
 int i;
 int y1,y2,x1,x2;
 
 /* Draw the box */
 drawbox (xwin, originx+TOUCHBORDER, originy+TOUCHBORDER, originx+WIDTH/XSEGMENTS-2*TOUCHBORDER, originy+TOUCHBORDER+8, RGB24RGB16(64,64,128));
 drawstring (xwin, originx+TOUCHBORDER, originy+TOUCHBORDER, label, WHITE);
 drawstring (xwin, originx+TOUCHBORDER, originy+TOUCHBORDER+TOUCHMODEROW*8, "Mode:", WHITE);
 drawstring (xwin, originx+TOUCHBORDER, originy+TOUCHBORDER+TOUCHSELROW*8, "Sel.:", WHITE);
 x1=originx+TOUCHBORDER+TOUCHCENTERX;
 x2=x1+3*TOUCHSEGMENTX;
 y1=originy+2*TOUCHBORDER+8*TOUCHTEXTLINES;
 y2=y1;
 hline(xwin,x1,x2,y2,WHITE);
 y2 += TOUCHSEGMENTY;
 hline(xwin,x1,x2,y2,WHITE);
 y2 += TOUCHSEGMENTY;
 hline(xwin,x1,x2,y2,WHITE);
 vline(xwin,x1,y1,y2,WHITE);
 x1 += TOUCHSEGMENTX;
 vline(xwin,x1,y1,y2,WHITE);
 x1 += TOUCHSEGMENTX;
 vline(xwin,x1,y1,y2,WHITE);
 x1 += TOUCHSEGMENTX;
 vline(xwin,x1,y1,y2,WHITE);
 /* Initialize centers */
 for (i=0; i<NUMTOUCH; i++) {
  touchcenterx[i]=originx+TOUCHBORDER+TOUCHCENTERX+TOUCHSEGMENTX/2+(i % 3)*TOUCHSEGMENTX;
  touchcentery[i]=originy+2*TOUCHBORDER+8*TOUCHTEXTLINES+TOUCHSEGMENTY/2+(i/3)*TOUCHSEGMENTY;
 }
}

/* Draw title background */

void drawTitle(struct xwinbuffer *xwin, char *label, int  originx, int originy)
{
 int i;
 char aline[128];
  
 drawbox (xwin, originx+JOYBORDER, originy+JOYBORDER, originx+WIDTH/XSEGMENTS-2*JOYBORDER, originy+JOYBORDER+8, RGB24RGB16(64,128,64));
 drawstring (xwin, originx+TITLEBORDER, originy+TITLEBORDER, label, WHITE);
 drawstring (xwin, originx+TITLEBORDER, originy+TITLEBORDER+2*8, copyright, WHITE);
 drawstring (xwin, originx+TITLEBORDER, originy+TITLEBORDER+3*8, dateversion, WHITE);
 drawstring (xwin, originx+TITLEBORDER, originy+TITLEBORDER+FPSROW*8, "FPS:", WHITE);
 drawstring (xwin, originx+TITLEBORDER, originy+TITLEBORDER+LIGHTROW*8, "Light:", WHITE);
 for (i=0; i<NUMDVD; i++) {
  sprintf(aline,"DVD %i playing:",i+1);
  drawstring (xwin, originx+TITLEBORDER, originy+TITLEBORDER+(CHAPTERROW+i)*8, aline, WHITE);
 }
}

/* Timing routines */

static double lasttime=0.0;
static double frametime=0.0;
static float fps=0.0;

/* Routine to get the time as float */
double timeGet(void)
{
  struct timeval tv;
  
  /* get wallclock time */
  gettimeofday(&tv, NULL);
  return((double)tv.tv_sec+1.0e-6*(double)tv.tv_usec);
}

/* Call for each frame to get FPS calculation */
void calc_fps()
{
 double curtime;
 
 curtime=timeGet();
 frametime=(curtime-lasttime);
 fps=(float)(1.0/frametime);
 lasttime=curtime;
}

/* ================= MAIN PROGRAM LOOP =================== */


int main (int argc, char *argv[])
{
 int i,j;
 int framecount=0;
 char aline[256];
 struct xwinbuffer xwin;
 int inloop=1;
 KeySym key;
 int button;
 int mousex,mousey;
 int selectedsegment=-1;
 unsigned short *background;
 float delaytime=0.0;
 struct adcreader adc[NUMADC];
 struct joyreader joy[NUMJOY];
 struct dvdcontroller dvd[NUMDVD];
 struct touchscanner touch;
 struct lightcontroller light;
 float localx, localy;
 int joyoffsetx, joyoffsety;
 int screenx, screeny;
 int segx1,segy1,segx2,segy2;
 int localtestmode;
 int selectedjoy=-1;
 int joysegment[NUMJOY];
 int joysubsegment[NUMJOY];
 int oldjoysegment[NUMJOY];
 int oldjoysubsegment[NUMJOY];
 int local_sensor_matrix[NUMTOUCH][NUMTOUCH];
 int touchsegment; 
 int oldtouchsegment;
 int firsttouch;
 int firsttouchsegment1;
 int firsttouchsegment2;
 FILE *iofile;
 double currenttime;
 int wantchapter;
 int conversationplaying;
 double conversationtimeout;
    
 /* ------------ Startup --------------- */

 /* Load setup information */
 LoadMappingMatrix();
 LoadConversationMatrix();
 
 /* Create communication threads */

 /* Analog/Digital Converters */
 fprintf (stdout,"Starting ADC interfaces ...\n");
 fflush (stdout);
 for (i=0; i<NUMADC; i++) {
  if (init_adcreader(&adc[i],adcdevice[i],adctype[i])==-1) {
   fprintf(stdout,"Failed to initialize AD converter %i on device %s - disabling it.\n",i,adcdevice[i]);
   fflush (stdout);
  } else {
   if (start_adcreader(&adc[i])==-1) {
    fprintf(stdout,"Failed to start AD converter %i.\n",i);
    fflush (stdout);
   }
  }
 }

 /* Joysticks */
 fprintf (stdout,"Starting Joystick readers ...\n");
 fflush (stdout);
 for (i=0; i<NUMJOY; i++) {
  joysegment[i]=-1;
  oldjoysegment[i]=-1;
  joysubsegment[i]=-1;
  oldjoysubsegment[i]=-1;
  fprintf (stdout,"Joystick %i: ADC %i X-Channel %i Y-Channel %i\n",i,i/(NUM_CHANNELS/2),2*(i % (NUM_CHANNELS/2)),2*(i % (NUM_CHANNELS/2))+1);
  fflush (stdout);
  if (init_joyreader(&joy[i],&adc[i/(NUM_CHANNELS/2)],2*(i % (NUM_CHANNELS/2)),2*(i % (NUM_CHANNELS/2))+1)==-1) {
   fprintf(stdout,"Failed to initialize joystick reader %i on ADC %i - disabling it.\n",i,i/(NUM_CHANNELS/2));
   fflush (stdout);
  } else {
   if (start_joyreader(&joy[i])==-1) {
    fprintf(stdout,"Failed to start joystick reader %i.\n",i);
    fflush (stdout);
   }
  }
 }
 
 /* Read joystick calibration data */
 if ((iofile=fopen (calibration_filename,"r"))!=NULL) {
  fprintf(stdout,"Reading joystick calibration data from '%s'.\n",calibration_filename);
  fflush (stdout);
  for (i=0; i<NUMJOY; i++) {
   joyreader_lock(&joy[i]);
   fscanf (iofile,"%f %f %f %f %f %f\n",&joy[i].minu,&joy[i].maxu,&joy[i].minv,&joy[i].maxv,&joy[i].xfactor,&joy[i].yfactor);
   fflush (stdout);
   joyreader_unlock(&joy[i]);
  }
  fclose(iofile);
 } else {
  fprintf(stdout,"Cannot read joystick calibration from '%s'.\n",calibration_filename);
  fflush (stdout);
 }  
 
 /* DVD players */
 fprintf (stdout,"Starting DVD interfaces ...\n");
 fflush (stdout);
 for (i=0; i<NUMDVD; i++) {
  if (init_dvdcontroller(&dvd[i],dvddevice[i],dvdtype[i])==-1) {
   fprintf(stdout,"Failed to initialize DVD controller %i on device %s - disabling it.\n",i,dvddevice[i]);
   fflush (stdout);
  } else {
   if (start_dvdcontroller(&dvd[i])==-1) {
    fprintf(stdout,"Failed to start %s-DVD controller %i.\n",dvd_names[dvd[i].type],i);
    fflush (stdout);
   } else {
    /* Send initializing commands */
    DVDCommand_Init(&dvd[i],i+1);
   }
  }
 }

 /* Touch interface */
 fprintf (stdout,"Starting touch interface ...\n");
 fflush (stdout);
 if (init_touchscanner(&touch,touchdevice,touchtype)==-1) {
  fprintf(stdout,"Failed to initialize touch interface - disabling it.\n");
  fflush (stdout);
 } else {
  if (start_touchscanner(&touch)==-1) {
   fprintf(stdout,"Failed to start touch interface on device %s\n.",touchdevice);
   fflush (stdout);
  }
 }
 for (i=0; i<NUMTOUCH; i++) {
  manualtouchstate[i]=0;
 }
 touchsegment=-1;
 oldtouchsegment=-1;
 firsttouch=-1;
 firsttouchsegment1=0;
 firsttouchsegment2=0;
 conversationplaying=0;
 conversationtimeout=0.0;

 /* Light controller */
 fprintf (stdout,"Starting light controller interface ...\n");
 fflush (stdout);
 if (init_lightcontroller(&light,lightdevice,lighttype)==-1) {
  fprintf(stdout,"Failed to initialize light controller - disabling it.\n");
  fflush (stdout);
 } else {
  if (start_lightcontroller(&light)==-1) {
   fprintf(stdout,"Failed to start light controller on device %s\n.",lightdevice);
   fflush (stdout);
  }
 }

 /* Create background buffer */
 if ((background=malloc(WIDTH*HEIGHT*DEPTH))==NULL) {
  perror("malloc()");
  exit(-1);
 }

 /* Create window */  
 if (setup_xwinbuffer (&xwin, argv[0], WIDTH, HEIGHT, DEPTH)!=0) {
  exit(-1);
 }

 /* Draw and backstore background */
 memset((char *)xwin.shmimage->data,0,SIZE);
 /* Title */
 drawTitle(&xwin,"Beyond Hierarchy",0,0);
 /* Touchbox */
 drawTouchbox(&xwin,"Touchbox",0,HEIGHT/YSEGMENTS);
 touchoriginx=0;
 touchoriginy=(HEIGHT/YSEGMENTS);
 /* Joystick boxes */
 for (i=1; i<XSEGMENTS; i++) {
  for (j=0; j<YSEGMENTS; j++) {
   /* Title name */
   sprintf(aline,"Joystick %i: %s",i+3*j,joyname[i+3*j-1]);
   /* Reference coordinate arrays */
   joyoriginx[(i-1)+j*3]=i*(WIDTH/XSEGMENTS);
   joyoriginy[(i-1)+j*3]=j*(HEIGHT/YSEGMENTS);
   joycenterx[(i-1)+j*3]=i*(WIDTH/XSEGMENTS)+((JOYSEGMENTS*JOYSEGMENTX)/2+JOYBORDER+JOYCENTERX);
   joycentery[(i-1)+j*3]=j*(HEIGHT/YSEGMENTS)+(((JOYSEGMENTS+1)*JOYSEGMENTY)/2+JOYBORDER+8*JOYTEXTLINES);
   joysizex[(i-1)+j*3]=JOYSEGMENTS*JOYSEGMENTX;
   joysizey[(i-1)+j*3]=JOYSEGMENTS*JOYSEGMENTY;
   /* Draw background */
   drawJoystickH(&xwin,aline,i*(WIDTH/XSEGMENTS),j*(HEIGHT/YSEGMENTS));
  } 
 }
 memcpy((char *)background, (char *)xwin.shmimage->data, SIZE);

 /* Output quit command */
 fprintf (stdout,"\nProgramm running - press 'Q' to quit.\n\n");
 fflush (stdout);
       
 /* -------------- View/Processing loop ---------------- */
 while (inloop) { 
  /* Calculate FPS, update framecount */
  calc_fps();
  framecount++;

  /* Copy background to screenbuffer */
  memcpy((char *)xwin.shmimage->data, (char *)background, SIZE);

  /* Show selectedsegment */
  if (selectedsegment>-1) {
   segx1=(selectedsegment % XSEGMENTS)*(WIDTH/XSEGMENTS);
   segy1=(selectedsegment / XSEGMENTS)*(HEIGHT/YSEGMENTS);
   segx2=segx1+(WIDTH/XSEGMENTS);
   segy2=segy1+(HEIGHT/YSEGMENTS);
   hline(&xwin,segx1,segx2,segy1,WHITE);
   hline(&xwin,segx1,segx2,segy2,WHITE);
   vline(&xwin,segx1,segy1,segy2,WHITE);
   vline(&xwin,segx2,segy1,segy2,WHITE);
  }

  /* --------- TITLE PROCESSING */
  /* Show FPS */
  sprintf(aline,"%4.0f",fps);
  drawstring (&xwin, TITLEBORDER+FPSCOL*8, TITLEBORDER+FPSROW*8, aline, RGB24RGB16(200,255,200) );
  sprintf(aline,"%3i",lightstate);
  drawstring (&xwin, TITLEBORDER+LIGHTCOLUMN*8, TITLEBORDER+LIGHTROW*8, aline, RGB24RGB16(200,255,200) );
  for (i=0; i<NUMDVD; i++) {
   sprintf(aline,"%2i",dvdchapter[i]);
   drawstring (&xwin, TITLEBORDER+CHAPTERCOLUMN*8, TITLEBORDER+(CHAPTERROW+i)*8, aline, RGB24RGB16(200,255,200));
  }

  /* --------- TOUCHPAD PROCESSING */
  /* get testmode */
  touchscanner_lock(&touch);
  localtestmode=touch.testmode;
  touchscanner_unlock(&touch);
  if (localtestmode==TOUCH_MANUAL) {
   strcpy (aline,"Manual");
   /* Create local connection matrix based on mouse input */
   memset(local_sensor_matrix,0,sizeof(sizeof(int)*NUMTOUCH*NUMTOUCH));
   for (i=0; i<NUMTOUCH; i++) {
    for (j=0; j<NUMTOUCH; j++) {
     if (i!=j) {
      local_sensor_matrix[i][j]=((manualtouchstate[i]==1) && (manualtouchstate[j]==1));
     }
    }
   }
  } else {
   strcpy (aline,"Normal");
   /* Copy connection matrix from touchscanner */
   touchscanner_lock(&touch);
   for (i=0; i<NUMTOUCH; i++) {
    manualtouchstate[i]=0;
    for (j=0; j<NUMTOUCH; j++) {
     local_sensor_matrix[i][j]=touch.sensor_matrix[i][j];
     if (local_sensor_matrix[i][j]) {
      manualtouchstate[i] |= 1;
      manualtouchstate[j] |= 1;
     } 
    }
   }
   touchscanner_unlock(&touch);
  }  
  drawstring(&xwin,touchoriginx+TOUCHBORDER+TOUCHMODECOL*8,touchoriginy+TOUCHBORDER+TOUCHMODEROW*8,aline,RGB24RGB16(200,200,255));
  if (conversationplaying) {
   wantchapter=conversation_matrix[firsttouchsegment1][firsttouchsegment2];
   sprintf (aline,"Playing %i",wantchapter);
  } else {
   sprintf (aline,"Waiting");
  } 
  drawstring(&xwin,touchoriginx+TOUCHBORDER+TOUCHSELCOL*8,touchoriginy+TOUCHBORDER+TOUCHSELROW*8,aline,RGB24RGB16(200,200,255));
  /* Draw manual sensor activations */
  for (i=0; i<NUMTOUCH; i++) {
   if (manualtouchstate[i]) {
    hline(&xwin,touchcenterx[i]-5,touchcenterx[i]+5,touchcentery[i]-5,WHITE);
    hline(&xwin,touchcenterx[i]-5,touchcenterx[i]+5,touchcentery[i]+5,WHITE);
    vline(&xwin,touchcenterx[i]-5,touchcentery[i]-5,touchcentery[i]+5,WHITE);
    vline(&xwin,touchcenterx[i]+5,touchcentery[i]-5,touchcentery[i]+5,WHITE);
   }
  }
  /* Draw sensor connections - do some processing */
  for (i=0; i<NUMTOUCH; i++) {
   if (local_sensor_matrix[i][0]) {
    hline(&xwin,touchcenterx[i]-5,touchcenterx[i]+5,touchcentery[i]-5,WHITE);
    hline(&xwin,touchcenterx[i]-5,touchcenterx[i]+5,touchcentery[i]+5,WHITE);
    vline(&xwin,touchcenterx[i]-5,touchcentery[i]-5,touchcentery[i]+5,WHITE);
    vline(&xwin,touchcenterx[i]+5,touchcentery[i]-5,touchcentery[i]+5,WHITE);
   }
   /* Only process when not playing a movie */   
   if (conversationplaying==0) {
    lockjoystick[i]=0;
    for (j=0; j<NUMTOUCH; j++) {
     if (i!=j) {
      if (local_sensor_matrix[i][j]==1) {
       /* Indicate connection */
       drawline(&xwin,touchcenterx[i],touchcentery[i],touchcenterx[j],touchcentery[j],RGB24RGB16(200,200,255));
       /* Put this joystick pair in 'locked' mode */
       lockjoystick[i]=1;
       lockjoystick[j]=1;
       /* Remember segments */
       firsttouchsegment1=i;
       firsttouchsegment2=j;
       /* ------- Play DVD chapter according to touch panel state */
       /* Check DVD command timeout */
       currenttime=timeGet();
       if ((currenttime-dvdcommandtime[6])>DVDCOMMANDTIMEOUT) {
        /* ... passed, can issue new command */
        wantchapter=conversation_matrix[firsttouchsegment1][firsttouchsegment2];
        if (wantchapter!=dvdchapter[6]) {
         dvdcommandtime[6]=currenttime;
         dvdchapter[6]=wantchapter;
         DVDCommand_Play(&dvd[6],wantchapter);
         conversationplaying=1;
         conversationtimeout=currenttime+conversation_time[firsttouchsegment1][firsttouchsegment2];
         /* Exit loop */
         j=NUMTOUCH;
        }
       }
      }
     }
    }
   }
  }
  
  /* DVD control while playing conversation */
  if (conversationplaying) {
   /* Check conversation timeout */
   currenttime=timeGet();
   if (currenttime>conversationtimeout) {
    if ((currenttime-dvdcommandtime[6])>DVDCOMMANDTIMEOUT) {
     /* Show black still frame */
     wantchapter=16;
     DVDCommand_Play(&dvd[6],wantchapter);
     dvdcommandtime[6]=currenttime;
     dvdchapter[6]=wantchapter;
     conversationplaying=0;
    }
   } else {
    /* Indicate conversation connection in WHITE */
    drawline(&xwin,touchcenterx[firsttouchsegment1],touchcentery[firsttouchsegment1],touchcenterx[firsttouchsegment2],touchcentery[firsttouchsegment2],RGB24RGB16(255,255,255));
   }
  }
  
  /* Initially set black frame on conversation */
  if (dvdchapter[6]==-1) {
   /* Show black still frame */
   currenttime=timeGet();
   wantchapter=16;
   DVDCommand_Play(&dvd[6],wantchapter);
   dvdcommandtime[6]=currenttime;
   dvdchapter[6]=wantchapter;
   conversationplaying=0;
  }
 
  /* --------- LIGHTS PROCESSING */
  lightcontroller_lock(&light);
  if (conversationplaying) {
   light.value = (touchlight[firsttouchsegment1]+touchlight[firsttouchsegment2]);
   lightstate=light.value;
  } else {
   light.value = 1;
   lightstate=light.value;
  }
  lightcontroller_unlock(&light);
    
  /* --------- JOYSTICK PROCESSING */
  /* Get joystick coordinates, map to trigger zones, act, display */

  for (i=0; i<NUMJOY; i++) {
   /* Get current coordinates */
   joyreader_lock(&joy[i]);
   localx=joy[i].x;
   localy=joy[i].y;
   localtestmode=joy[i].testmode;
   joyreader_unlock(&joy[i]);
   /* Show testmode */
   if (localtestmode==JOY_MANUAL) {
    strcpy (aline,"Manual");
   } else if (localtestmode==JOY_CALIBRATE) {
    strcpy (aline,"Calibrate");
   } else {
    strcpy (aline,"Normal");
   }
   drawstring(&xwin,joyoriginx[i]+JOYBORDER+JOYMODECOL*8,joyoriginy[i]+JOYBORDER+JOYMODEROW*8,aline,RGB24RGB16(255,200,200));
   /* Show locking */
   if (lockjoystick[i]) {
    strcpy (aline,"Locked");
    /* Blinking */
    if ((framecount % (int)TARGETFPS)<((int)TARGETFPS/2)) {
     drawstring(&xwin,joyoriginx[i]+JOYBORDER+JOYLOCKCOL*8,joyoriginy[i]+JOYBORDER+JOYLOCKROW*8,aline,RGB24RGB16(255,200,200));
    } else {
     drawstring(&xwin,joyoriginx[i]+JOYBORDER+JOYLOCKCOL*8,joyoriginy[i]+JOYBORDER+JOYLOCKROW*8,aline,WHITE);
    }
   } else {
    strcpy (aline,"Off");
    drawstring(&xwin,joyoriginx[i]+JOYBORDER+JOYLOCKCOL*8,joyoriginy[i]+JOYBORDER+JOYLOCKROW*8,aline,RGB24RGB16(255,200,200));
   }

   /* Calculate joystick active area */
   joyoffsetx=(int)(0.5*(localx+1.0)*(float)JOYSEGMENTS);
   if (joyoffsetx==JOYSEGMENTS) {
    joyoffsetx=(JOYSEGMENTS-1);
   }
   joyoffsety=(int)(0.5*(localy+1.0)*(float)JOYSEGMENTS);
   if (joyoffsety==JOYSEGMENTS) {
    joyoffsety=(JOYSEGMENTS-1);
   }
   joysegment[i]=joytarget[joyoffsety][joyoffsetx];
   if ((joysegment[i]== 2) || (joysegment[i]== 3) || (joysegment[i]== 5) || (joysegment[i]== 6)) {
    /* Subdivide along y, left */
    joyoffsety=(int)(0.5*(localy+1.0)*(float)(joysizey[i]-1));
    joysubsegment[i]=1 + ((joyoffsety-(joysegment[i]-1)*JOYSEGMENTY))/(JOYSEGMENTY/SUBDIVY);
   } else if ((joysegment[i]==14) || (joysegment[i]==15) || (joysegment[i]==17) || (joysegment[i]==18)) {
    /* Subdivide along y, right */
    joyoffsety=(int)(0.5*(localy+1.0)*(float)(joysizey[i]-1));
    joysubsegment[i]=1 + ((joyoffsety-(joysegment[i]-13)*JOYSEGMENTY))/(JOYSEGMENTY/SUBDIVY);
   } else if ((joysegment[i]==8) || (joysegment[i]==9) ||(joysegment[i]==11) ||(joysegment[i]==12)) {
    /* Subdivide along x */
    joyoffsetx=(int)(0.5*(localx+1.0)*(float)(joysizex[i]-1));
    joysubsegment[i]=1 + ((joyoffsetx-(joysegment[i]-7)*JOYSEGMENTX))/(JOYSEGMENTX/SUBDIVX);
   } else {
    /* Don't subdivide */
    joysubsegment[i] = 1;
   }
   /* Enlarge middle by remapping subsegments */
   if (joysegment[i]==9) {
    if (joysubsegment[i]>2) {
     joysegment[i]=10;
     joysubsegment[i]=1;
    }  
   }
   if (joysegment[i]==11) {
    if (joysubsegment[i]<3) {
     joysegment[i]=10;
     joysubsegment[i]=1;
    }  
   }
   /* Draw box to indicate location */
   if (joysegment[i]!=-1) {
    drawbox(&xwin, joyoriginx[i]+  JOYBORDER+1+JOYCENTERX                  +joytargetx[joysegment[i]-1]*JOYSEGMENTX,
                  joyoriginy[i]+2*JOYBORDER+1+8*JOYTEXTLINES               +joytargety[joysegment[i]-1]*JOYSEGMENTY,
                  joyoriginx[i]+  JOYBORDER+1+JOYCENTERX+JOYSEGMENTX-1     +joytargetx[joysegment[i]-1]*JOYSEGMENTX,
                  joyoriginy[i]+2*JOYBORDER+1+8*JOYTEXTLINES+JOYSEGMENTY-2 +joytargety[joysegment[i]-1]*JOYSEGMENTY,
                  RGB24RGB16(128,128,128));
   }
       
   /* Print current segment */
   sprintf(aline,"%2i %2i",joysegment[i],joysubsegment[i]);
   drawstring(&xwin,joyoriginx[i]+JOYBORDER+JOYSELCOL*8,joyoriginy[i]+JOYBORDER+JOYSELROW*8,aline,RGB24RGB16(255,200,200));

   /* ------- Play DVD chapter according to joystick position*/
   /* Check DVD command timeout */
   currenttime=timeGet();
   if ((currenttime-dvdcommandtime[i])>DVDCOMMANDTIMEOUT) {
    /* ... passed, can issue new command */
    if (lockjoystick[i]) {
     wantchapter=mapping_matrix[i][joysegment[i]-1][joysubsegment[i]-1][1];
    } else {
     wantchapter=mapping_matrix[i][joysegment[i]-1][joysubsegment[i]-1][0];
    }
    if (wantchapter!=-1) {
     if (wantchapter!=dvdchapter[i]) {
      dvdcommandtime[i]=currenttime;
      dvdchapter[i]=wantchapter;
      DVDCommand_Play(&dvd[i],wantchapter);
     }
    }
   }

   /* Display coordinate with a little cross */
   joyoffsetx=(int)(0.5*(localx+1.0)*(float)joysizex[i]);
   joyoffsety=(int)(0.5*(localy+1.0)*(float)joysizey[i]);
   joyoffsetx=(int)(0.5*localx*(float)joysizex[i]);
   screenx=joycenterx[i]+joyoffsetx;
   joyoffsety=(int)(0.5*localy*(float)joysizey[i]);
   screeny=joycentery[i]+joyoffsety;
   vline(&xwin,screenx,screeny-5,screeny+5,RGB24RGB16(255,127,127));
   hline(&xwin,screenx-5,screenx+5,screeny,RGB24RGB16(255,127,127));
  
  } /* ... loop over all joysticks */
 
  
  /* Draw XWindow buffer */
  sync_xwinbuffer(&xwin);
 
  /* ----------- User interface ---------- */

  /* Poll keyboard */
  button=0;
  mousex=0;
  mousey=0;
  if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {
   switch (key) {
    case XK_q:
     /* Quit program */
     inloop=0;
     break; 
    case XK_s: 
     /* Save joystick calibration */
     if ((iofile=fopen (calibration_filename,"w"))!=NULL) {
      fprintf(stdout,"Saving joystick calibration data ...\n");
      fflush (stdout);
      for (i=0; i<NUMJOY; i++) {
       joyreader_lock(&joy[i]);
       fprintf(iofile,"%f %f %f %f %f %f\n",joy[i].minu,joy[i].maxu,joy[i].minv,joy[i].maxv,joy[i].xfactor,joy[i].yfactor); 
       joyreader_unlock(&joy[i]);
      }
      fclose(iofile);
     }
     break;
    case XK_c:
     /* Joystick calibration mode */
     if (selectedjoy>-1) {
      joyreader_lock(&joy[selectedjoy]);
      joy[selectedjoy].testmode=JOY_CALIBRATE;
      joy[selectedjoy].minu=1.0;
      joy[selectedjoy].maxu=0.0;
      joy[selectedjoy].minv=1.0;
      joy[selectedjoy].maxv=0.0;
      joyreader_unlock(&joy[selectedjoy]);
     }
     break; 
    case XK_m:
     /* Joystick manual mode */
     if (selectedjoy>-1) {
      joyreader_lock(&joy[selectedjoy]);
      joy[selectedjoy].testmode=JOY_MANUAL;
      joyreader_unlock(&joy[selectedjoy]);
     }
     /* Touchpad manual mode */
     if (selectedsegment==XSEGMENTS) {
      touchscanner_lock(&touch);
      touch.testmode=TOUCH_MANUAL;
      touchscanner_unlock(&touch);
     }
     break; 
    case XK_n:
     /* Joystick normal processing mode */
     if (selectedjoy>-1) {
      joyreader_lock(&joy[selectedjoy]);
      joy[selectedjoy].testmode=JOY_NORMAL;
      joyreader_unlock(&joy[selectedjoy]);
     }
     /* Touchpad normal mode */
     if (selectedsegment==XSEGMENTS) {
      touchscanner_lock(&touch);
      touch.testmode=TOUCH_NORMAL;
      touchscanner_unlock(&touch);
     }
     break; 
    case XK_Left:
     /* Move selection left */
     if ((selectedsegment>0) && (selectedsegment!=XSEGMENTS)) { 
      selectedsegment--;
     }
     if (selectedsegment==0) selectedsegment=-1;
     /* Determine selected joystick */     
     selectedjoy=-1;
     if ((selectedsegment>0) && (selectedsegment<XSEGMENTS)) {
      selectedjoy=selectedsegment-1;
     } else if ((selectedsegment>XSEGMENTS) && (selectedsegment<(XSEGMENTS*YSEGMENTS))) {
      selectedjoy=selectedsegment-2;
     }
     break; 
    case XK_Right:
     /* Move selection right */
     if ((selectedsegment<(YSEGMENTS*XSEGMENTS-1)) && (selectedsegment!=(XSEGMENTS-1)) ) { 
      selectedsegment++;
     }
     if (selectedsegment==0) selectedsegment=1;
     /* Determine selected joystick */     
     selectedjoy=-1;
     if ((selectedsegment>0) && (selectedsegment<XSEGMENTS)) {
      selectedjoy=selectedsegment-1;
     } else if ((selectedsegment>XSEGMENTS) && (selectedsegment<(XSEGMENTS*YSEGMENTS))) {
      selectedjoy=selectedsegment-2;
     }
     break; 
    case XK_Up:
     /* Move selection up */
     if (selectedsegment>(XSEGMENTS-1)) { 
      selectedsegment -= XSEGMENTS;
     }
     if (selectedsegment==0) selectedsegment=-1;
     /* Determine selected joystick */     
     selectedjoy=-1;
     if ((selectedsegment>0) && (selectedsegment<XSEGMENTS)) {
      selectedjoy=selectedsegment-1;
     } else if ((selectedsegment>XSEGMENTS) && (selectedsegment<(XSEGMENTS*YSEGMENTS))) {
      selectedjoy=selectedsegment-2;
     }
     break; 
    case XK_Down:
     /* Move selection down */
     if (selectedsegment<XSEGMENTS) { 
      if (selectedsegment==-1) selectedsegment=0;
      selectedsegment += 4;
     }
     /* Determine selected joystick */     
     selectedjoy=-1;
     if ((selectedsegment>0) && (selectedsegment<XSEGMENTS)) {
      selectedjoy=selectedsegment-1;
     } else if ((selectedsegment>XSEGMENTS) && (selectedsegment<(XSEGMENTS*YSEGMENTS))) {
      selectedjoy=selectedsegment-2;
     }
     break; 
   } // switch

   /* Handle mouse clicks */
   if (button) {
    /* Determine selected segment */
    selectedsegment=(mousex/(WIDTH/XSEGMENTS))+XSEGMENTS*(mousey/(HEIGHT/YSEGMENTS));
    if (selectedsegment==0) selectedsegment=-1;
    /* Check touchpad */
    if (selectedsegment==XSEGMENTS) {
     i=(mousex-touchoriginx-TOUCHBORDER-TOUCHCENTERX)/TOUCHSEGMENTX;
     j=(mousey-touchoriginy-2*TOUCHBORDER-8*TOUCHTEXTLINES)/TOUCHSEGMENTY;
     if ((i>=0) && (i<3) && (j>=0) && (j<2)) {
      i += 3*j;
      manualtouchstate[i] = ((manualtouchstate[i]+1) & 1);
     }
    } 
    /* Determine selected joystick */     
    selectedjoy=-1;
    if ((selectedsegment>0) && (selectedsegment<XSEGMENTS)) {
     selectedjoy=selectedsegment-1;
    } else if ((selectedsegment>XSEGMENTS) && (selectedsegment<(XSEGMENTS*YSEGMENTS))) {
     selectedjoy=selectedsegment-2;
    }
    /* Get testmode if joystick is selected */
    if (selectedjoy>-1) {
     /* Get current testmode */
     joyreader_lock(&joy[selectedjoy]);
     localtestmode=joy[selectedjoy].testmode;
     if (localtestmode==JOY_MANUAL) {
      /* Set coordinates from mouse input */
      localx=(float)(mousex-joycenterx[selectedjoy])/(0.5*(float)joysizex[selectedjoy]);
      localy=(float)(mousey-joycentery[selectedjoy])/(0.5*(float)joysizey[selectedjoy]);
      joy[selectedjoy].x=localx;
      joy[selectedjoy].y=localy;     
     }
     joyreader_unlock(&joy[selectedjoy]);
    }
   }
  }

  /* Insert pause to slow to target FPS */
  if (frametime<1.0) {
   delaytime=(float)(0.5*(delaytime+(1000000.0/TARGETFPS)-(1000000.0*frametime)));
   if (delaytime>1000.0) {
    usleep((unsigned long)delaytime);
   }
  }
  
 } // if inloop           

 /* Close window and clean up threads */
 if (close_xwinbuffer(&xwin)!=0) exit(-1);
 free (background);

 /* Light controller */
 fprintf (stdout,"Stopping light controller ...\n");
 if (stop_lightcontroller(&light)==-1) {
  fprintf(stdout,"Failed to stop light controller\n.");
 }

 /* Touch interface */
 fprintf (stdout,"Stopping touch interface ...\n");
 if (stop_touchscanner(&touch)==-1) {
  fprintf(stdout,"Failed to stop touch interface\n.");
 }

 /* DVD players */
 fprintf (stdout,"Stopping DVD interfaces ...\n");
 for (i=0; i<NUMDVD; i++) {
  if (stop_dvdcontroller(&dvd[i])==-1) {
   fprintf(stdout,"Failed to stop DVD controller %i.\n",i);
  }
 }

 /* Joysticks */
 fprintf (stdout,"Stopping joystick readers ...\n");
 for (i=0; i<NUMJOY; i++) {
  if (stop_joyreader(&joy[i])==-1) {
   fprintf(stdout,"Failed to stop joystick reader %i.\n",i);
  }
 }

 /* ADCs */
 fprintf (stdout,"Stopping ADC interfaces ...\n");
 for (i=0; i<NUMADC; i++) {
  if (stop_adcreader(&adc[i])==-1) {
   fprintf(stdout,"Failed to stop ADC reader %i.\n",i);
  }
 }
 fflush (stdout);

 exit(0);
}
