#
#include "touchscan.h"

/* Touch-Interface scanner */

int ParseInterfaceResult(struct touchscanner *touch, char *line)
{
 /* Determine input state */
 /* Check for "I" character */
 if (line[0]==(char)73) {
  /* The bit is set if 19th character is one of 8,9,A,B,C,D,E,F */
  if ((line[19]==(char)56) || (line[19]==(char)57) || ((line[19]>=(char)65) && (line[19]<=(char)70))) {
   return(1); 
  } else { 
   return(0); 
  } 
 } else {
  return(0);
 }
}

int ScannerThread(struct touchscanner *touch)
{
 int i;
 int inloop;
 int fd;
 int result;
 char buffer[SERBUFFERLEN];
 char *curbuffer;
 char linebuffer[SERBUFFERLEN];
 char *curlinebuffer;
 int linebufferpos;
 struct timeval timeout;
 int maxfd;
 fd_set readfs;
 int commandcounter;
 int previouscommandcounter;
 int readlines;
 int state;
 int testmode;

 /* Get local filehandle */
 touchscanner_lock(touch);
 fd=touch->fd;
 touchscanner_unlock(touch);

 /* Dummy read */
 FD_SET(fd, &readfs);
 timeout.tv_usec = 10;  /* milliseconds */       
 timeout.tv_sec  = 0;   /* seconds */       
 result = select(maxfd, &readfs, NULL, NULL, &timeout);
 if (result>0) {
  result = read(fd,buffer,SERBUFFERLEN);
 }
 
 /* Serial-port processing loop */
 inloop=1;
 commandcounter=0;
 previouscommandcounter=0;
 maxfd=fd+1;
 linebufferpos=0;
 curlinebuffer=linebuffer;
 readlines=0;
 while (inloop) {
  FD_SET(fd, &readfs);
  /* set timeout value within input loop */
  timeout.tv_usec = 10;  /* milliseconds */       
  timeout.tv_sec  = 0;   /* seconds */       
  /* Wait for data */
  result = select(maxfd, &readfs, NULL, NULL, &timeout);
  if (result>0) {
   /* ---- READ section ----- */
   /* Have data, read it */
   result = read(fd,buffer,SERBUFFERLEN);
   buffer[result]=0;
   /* Transfer data into linebuffer */
   curbuffer=buffer;
   for (i=0; i<result; i++) {
    /* Process buffer */
    if (*curbuffer=='\r') {
     /* Line is done */
     *curlinebuffer=0;
     if ((commandmatrix[previouscommandcounter].elementA)!=-1) {
      /* Get testmode */
      touchscanner_lock(touch);
      testmode=touch->testmode;
      touchscanner_unlock(touch);
      if (testmode==TOUCH_NORMAL) {
       /* Process content */
       state=ParseInterfaceResult(touch, linebuffer);
       /* Store state */
       touchscanner_lock(touch);
       touch->sensor_matrix[commandmatrix[previouscommandcounter].elementA][commandmatrix[previouscommandcounter].elementB]=state;
       touch->sensor_matrix[commandmatrix[previouscommandcounter].elementB][commandmatrix[previouscommandcounter].elementA]=state;
       touchscanner_unlock(touch);
      }
     }
     if (readlines>0) readlines--;
     /* Reset line */
     curlinebuffer=linebuffer;
     linebufferpos=0;     
    } else {
     /* Add character to line */
     *curlinebuffer=*curbuffer;
     curlinebuffer++;
     linebufferpos++;
    }
    /* Check line length */
    if (linebufferpos==SERBUFFERLEN) {
     /* Line is full */
     *curlinebuffer=0;
     if ((commandmatrix[previouscommandcounter].elementA)!=-1) {
      /* Get testmode */
      touchscanner_lock(touch);
      testmode=touch->testmode;
      touchscanner_unlock(touch);
      if (testmode==TOUCH_NORMAL) {
       /* Process content */
       state=ParseInterfaceResult(touch, linebuffer);
       /* Store state */
       touchscanner_lock(touch);
       touch->sensor_matrix[commandmatrix[previouscommandcounter].elementA][commandmatrix[previouscommandcounter].elementB]=state;
       touch->sensor_matrix[commandmatrix[previouscommandcounter].elementB][commandmatrix[previouscommandcounter].elementA]=state;
       touchscanner_unlock(touch);
      }
     }
     if (readlines>0) readlines--;
     /* Reset line */
     curlinebuffer=linebuffer;
     linebufferpos=0;     
    };
    curbuffer++;
   }
  }

  /* ---- WRITE section ---- */
  /* Send command - once all responses are in */
  if (readlines==0) {
   strcpy (buffer,commandmatrix[commandcounter].command);
   strcat (buffer,"\n\r");  
   result=write(fd,buffer,strlen(commandmatrix[commandcounter].command)+2);
   usleep(WRITEDELAY);
   /* Check for number of lines to be read */
   if (commandmatrix[commandcounter].responses!=0) {   
    /* Update readline counter */
    readlines += commandmatrix[commandcounter].responses;
   } 
   
   /* Update counter for next command */
   previouscommandcounter=commandcounter;
   commandcounter++;
   if (commandcounter==NUMCOMMANDS) {
    commandcounter=LOOPAT;
   }
  }
 }
 
 return(0);
}

int touchscanner_lock(struct touchscanner *touch)
{
 /* Lock shared variables */ 
 pthread_mutex_lock(&touch->touchscan_mutex);
 
 return(0);
}

int touchscanner_unlock(struct touchscanner *touch)
{
 /* Unlock shared variables */ 
 pthread_mutex_unlock(&touch->touchscan_mutex);
 
 return(0);
}

int start_touchscanner (struct touchscanner *touch)
{
 pthread_t reader_thread;

 /* Check if enabled */
 if (touch->enabled) {
  /* Store parameters */
  touchscanner_lock(touch);
  touchscanner_unlock(touch);
  /* Start thread */
  if (pthread_create(&reader_thread, NULL,(void *)ScannerThread,(void *)touch)==-1) {
   perror ("pthread_create()");
   return(-1);
  } else {
   touch->thread=reader_thread;
   return(0);
  }
 } else {
  fprintf (stderr,"start_touchscanner: cannot start thread - device not enabled.\n");
  return(-1);
 }
}

int init_touchscanner(struct touchscanner *touch, char *device, int type)
{
 int i,j;
 struct termios newtio;

 /* Turn off */
 touch->enabled=0;

 /* Initialize variables */
 for (i=0; i<NUM_SENSORS; i++) {
  for (j=0; j<NUM_SENSORS; j++) {
   touch->sensor_matrix[i][j]=0;
  }
 }
 touch->testmode=TOUCH_NORMAL;

 /* Initialize sycro mutex */
 if (pthread_mutex_init(&touch->touchscan_mutex, NULL)==-1) {
  perror("pthread_mutex_init()");
  return(-1);
 }

 /* Initialize variables */
 strncpy(touch->device,device,SERBUFFERLEN);

 if (type==TOUCH_NONE) {
  return(0);
 }

 /* Open device and set type */
 touch->fd = open(device, O_RDWR | O_NOCTTY );
 if (touch->fd<0) {
  perror("open()"); 
  return(-1); 
 }

 /* Termial setup */
 tcgetattr(touch->fd,&touch->oldtio); 
 bzero(&newtio, sizeof(newtio));
 newtio.c_cflag = B57600 | CRTSCTS | CS8 | CLOCAL | CREAD;
 newtio.c_iflag = IGNPAR;
 newtio.c_oflag = 0;
 newtio.c_lflag = 0;
 newtio.c_cc[VTIME]    = 0;
 newtio.c_cc[VMIN]     = 1;
 tcflush(touch->fd, TCIFLUSH);
 tcsetattr(touch->fd,TCSANOW,&newtio);
 
 /* Turn on */
 touch->enabled=1;
 return(0);
}

int stop_touchscanner (struct touchscanner *touch)
{
 if (touch->enabled) {
  /* Shutdown thread */
  if (touch->thread!=-1) {
   pthread_cancel(touch->thread);
  }

  /* Restore the old port settings */
  tcsetattr(touch->fd,TCSANOW,&touch->oldtio);

  /* Close device */
  close(touch->fd);
 }

 return(0); 
}
