
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>
#include <termios.h>

#include <errno.h>

#include <pthread.h> 

#define NUM_SENSORS	6

#define TOUCH_MANUAL	0
#define TOUCH_NORMAL	1

#define TOUCH_NONE	0
#define TOUCH_HAPP	1

#define SERBUFFERLEN	1024
#define WRITEDELAY	1000

#define NUMCOMMANDS	(6+2*15)
#define LOOPAT		6

struct happcommand {
			char command[SERBUFFERLEN];
			int responses;
			int elementA, elementB;
		};

/* TEST - turn on all outputs one-by-one */

static struct happcommand commandmatrixtest[NUMCOMMANDS] = {
					{ "R",4, -1,-1},
					{ "P0",1, -1,-1 },
					{ "M000000",0, -1,-1 },
					{ "S",1, -1,-1 },
					{ "X008000",0, -1,-1 },
					{ "T",1, -1,-1 },

					{ "O000001",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000002",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000004",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000008",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000010",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000020",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000040",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000080",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000100",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000200",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000400",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O000800",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O001000",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O002000",0, -1,-1 },
					{ "I",1, -1,-1 },

					{ "O004000",0, -1,-1 },
					{ "I",1, -1,-1 }

				};	

/* NORMAL sequence for first 6 inputs */

static struct happcommand commandmatrix[NUMCOMMANDS] = {
					{ "R",4, -1,-1 },
					{ "P0",1, -1,-1 },
					{ "M000000",0, -1,-1 },
					{ "S",1, -1,-1 },
					{ "X008000",0, -1,-1 },
					{ "T",1, -1,-1 },

					{ "O000101",0, -1,-1 },
					{ "I",1, 0,1 },

					{ "O000201",0, -1,-1 },
					{ "I",1, 0,2 },

					{ "O000401",0, -1,-1 },
					{ "I",1, 0,3 },

					{ "O000801",0, -1,-1 },
					{ "I",1, 0,4 },

					{ "O001001",0 },
					{ "I",1, 0,5 },

					{ "O000202",0, -1,-1 },
					{ "I",1, 1,2 },

					{ "O000402",0, -1,-1 },
					{ "I",1, 1,3 },

					{ "O000802",0, -1,-1 },
					{ "I",1, 1,4 },

					{ "O001002",0, -1,-1 },
					{ "I",1, 1,5 },

					{ "O000404",0, -1,-1 },
					{ "I",1, 2,3 },

					{ "O000804",0, -1,-1 },
					{ "I",1, 2,4 },

					{ "O001004",0, -1,-1 },
					{ "I",1, 2,5 },

					{ "O000808",0, -1,-1 },
					{ "I",1, 3,4 },

					{ "O001008",0, -1,-1 },
					{ "I",1, 3,5 },

					{ "O001010",0, -1,-1 },
					{ "I",1, 4,5 }
					
				};	
 
struct touchscanner {
 		 int enabled;
	         pthread_t thread;
		 pthread_mutex_t touchscan_mutex;
		 int testmode;
		 int sensor_matrix[NUM_SENSORS][NUM_SENSORS];
                 int fd;
                 char device[SERBUFFERLEN];
                 struct termios oldtio;
		};

/* Prototypes */

int init_touchscanner(struct touchscanner *touch, char *device, int type);
int start_touchscanner(struct touchscanner *touch); 
int stop_touchscanner (struct touchscanner *touch);

int touchscanner_lock(struct touchscanner *touch);
int touchscanner_unlock(struct touchscanner *touch);

